// IckarSocket.h: interface for the CIckarSocket class.
//
//////////////////////////////////////////////////////////////////////
#pragma once

#include "IIckarSocket.h"
#include "Mutex.h"
#if defined(WIN32)
#include "AsyncSocketEx.h"
#else
#include "AsyncSocketEx_bsdsocket.h"
#include <arpa/inet.h>
#include <errno.h>
#endif

#pragma warning( disable:4786 )

#include <list>
using namespace std;

#if defined(WIN32)
class CAsyncProxySocketLayer;
#endif

/**
  @ingroup NpSessionBase
  @brief CIkarSocket  ۽Ź۷ ϱ ť Ŭ
  @author 翵
*/
class ICKARSOCKET_SEND_QUEUE_DATA
{
public:
	ICKARSOCKET_SEND_QUEUE_DATA(LPCSTR szData, int nLen)
	{
		m_pData = new char[nLen+1];
		memcpy(m_pData, szData, nLen+1);
		m_nLen = nLen;
	}
	ICKARSOCKET_SEND_QUEUE_DATA(char* szDataBuff[], USHORT usDataLen[], USHORT count)
	{
		int nTotalLen = 0;
		for( int i=0 ; i<count ; i++ )
			nTotalLen += usDataLen[i];

		m_pData = new char[nTotalLen+1];
		int nWritten = 0;
		for( int i=0 ; i<count ; i++ )
		{
            memcpy(m_pData + nWritten, szDataBuff[i], usDataLen[i]);
			nWritten += usDataLen[i];
		}

		m_nLen = nWritten;
		ASSERT( nTotalLen == nWritten );
	}
	~ICKARSOCKET_SEND_QUEUE_DATA()
	{
		delete m_pData;
	}
	char * m_pData;
	int m_nLen;
};


/**
  @ingroup NpSessionBase
  @brief CAsyncSocketEx   Ŭ.
  @author 翵
*/
class CIckarSocket : public IIckarSocket
, public CAsyncSocketEx
{
public:
	CIckarSocket();
	virtual ~CIckarSocket();

// IIckarSocket implementation
	virtual _IIckarSocketEvents * SetEvents( _IIckarSocketEvents * pNewEvents );
	virtual bool Bind(int nBindPort, LPCSTR szBindAddr = NULL);
	virtual void Close();
#if defined(WIN32)
	virtual void ShutDown(E_NCS_SD_TYPE eType = eNCS_SD_SEND);
#endif
	virtual bool GetSockName(int & nSocketPort, char szSocketAddress[], int nMax = 16);
	virtual bool GetPeerName(int & nPeerPort, char szPeerAddress[], int nMax = 16);
	virtual bool Listen(int nConnectionBacklog);
	virtual bool Accept(IIckarSocket * pIckarSocket);
	virtual bool IsConnected();
	virtual int GetLastError();
	virtual bool GetLastWouldBlockError();

	virtual bool Connect(LPCSTR szHostAddr, int nHostPort, int nTimeOut = -1);
	virtual void Send(LPCSTR szData, int nLen);
	virtual int Receive(LPSTR szData, int nLen);

	virtual void Delete();

	virtual int ReceiveFrom(LPSTR lpBuf, int nBufLen,
		USHORT &uFromPort, char szFromIP[], int nMax = 16, int nFlags = 0);

	virtual int SendTo(LPCSTR szData, int nDataLen,
		USHORT uTargetPort, LPCSTR szTargetIP, int nFlags = 0);

	virtual bool ConnectEx(LPCSTR szHostAddr, int nHostPort, E_NCS_PROTOCOL_TYPE ePType = eNCS_PT_TCP, int nTimeOut = -1);
	virtual bool BindEx(int nBindPort, LPCSTR szBindAddr, E_NCS_PROTOCOL_TYPE ePType = eNCS_PT_TCP);
#if defined(WIN32)
	virtual void CheckSocketNotify();

	virtual bool AttachSocket(int socket);

	virtual bool SetRTSMode();

	virtual E_NCS_PROTOCOL_TYPE GetNcsProtocalType();
	virtual void Sendv(char* szDataBuff[], USHORT usDataLen[], USHORT count);
#endif
	virtual int DetachSocket();
	virtual int GetSocketHandle();

	virtual bool HasSendData()
	{
		return m_queueSendData.size() > 0;
	}
	
	virtual void SetMainThreadEventMode(bool bEnable)
	{
#if defined(WIN32)
#else
		CAsyncSocketEx::SetMainThreadEventMode(bEnable);
#endif
	}
	
	bool Check_AsyncDelete();

protected:
	void OnReceive(int nErrorCode, int nBytes);
	void OnSend(int nErrorCode);
	void OnConnect(int nErrorCode);
	void OnAccept(int nErrorCode);
	void OnClose(int nErrorCode);
	void OnSocketError(int nErrorCode);

	list< ICKARSOCKET_SEND_QUEUE_DATA * > m_queueSendData;
	
	bool m_bConnected;
	int m_nLastError;
	bool m_bSending;
	bool m_bDisconnected;

	void AppropriateSocketOption();

	_IIckarSocketEvents * m_pEvents;
#if defined(WIN32)
	CAsyncProxySocketLayer * m_pProxyLayer;
#endif
	bool m_bUseRTSMode;
	volatile bool m_bDeleteMark;
#if defined(WIN32)
	volatile LONG m_nFiringEventCount;
#else
	int  m_nFiringEventCount;	
#endif
	bool m_bNotifyClosed;
	
	nfc::CMutex m_mutexEvent;
	nfc::CMutex m_mutex;
};

inline
_IIckarSocketEvents * CIckarSocket::SetEvents( _IIckarSocketEvents * pNewEvents )
{
	_IIckarSocketEvents * pOldEvents;
	
	pOldEvents = m_pEvents;
	
	m_pEvents = pNewEvents;

	return pOldEvents;
}


inline
bool CIckarSocket::GetLastWouldBlockError()
{
#if defined(WIN32)
	if(GetLastError() == WSAEWOULDBLOCK)
		return true;
#else
	if(GetLastError() == EWOULDBLOCK)
		return true;
#endif
		
	return false;
}


inline 
int CIckarSocket::GetLastError()
{
#if defined(WIN32)
	if (m_nLastError)
	{
		int nError = m_nLastError;
		m_nLastError = 0;
		return nError;
	}
	
	return CAsyncSocketEx::GetLastError();
#else
	return CAsyncSocketEx::GetLastError();
#endif
}

inline 
void CIckarSocket::AppropriateSocketOption()
{
#if defined(WIN32) || defined(__USE_BSDSOCKET_WITH_KQUEUE__)
	int buff_len;
	socklen_t optlen;

	int result;

	optlen = sizeof(buff_len);
	result = getsockopt(m_hSocket, SOL_SOCKET, SO_SNDBUF, (char*)&buff_len, &optlen );
	if( result == SOCKET_ERROR )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
	}

	buff_len = 8192*2;
	result = setsockopt(m_hSocket, SOL_SOCKET, SO_SNDBUF, (char*)&buff_len, sizeof(buff_len));
	if( result == SOCKET_ERROR )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
	}

	optlen = sizeof(buff_len);
	result = getsockopt(m_hSocket, SOL_SOCKET, SO_SNDBUF, (char*)&buff_len, &optlen );
	if( result == SOCKET_ERROR )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
	}
#endif
}


inline
bool CIckarSocket::Check_AsyncDelete()
{
	if( m_bDeleteMark )
	{
		if( m_nFiringEventCount == 0 )
		{
			//DLOG( "[ENTER] CIckarSocket::Check_AsyncDelete() %p %d", this, m_hSocket);

			Delete();
			
			return true;
		}
	}

	return false;
}
